"""
The Raster GIS module contains functions for attributing and manipulating raster attribute tables.
"""

# import the C++ extension into this level
from ._rastergis import *

import osgeo.gdal as gdal

import numpy

haveHDF5 = True
try:
    import h5py
except ImportError as h5Err:
    haveHDF5 = False


class BandAttStats:
    """ This is passed to the populateRATWithStats function """
    def __init__(self, band, minField=None, maxField=None, sumField=None, stdDevField=None, meanField=None):
        self.band = band
        self.minField = minField
        self.maxField = maxField
        self.sumField = sumField
        self.meanField = meanField
        self.stdDevField = stdDevField
        
class FieldAttStats:
    """ This is passed to the calcRelDiffNeighStats function """
    def __init__(self, field, minField=None, maxField=None, sumField=None, stdDevField=None, meanField=None):
        self.field = field
        self.minField = minField
        self.maxField = maxField
        self.sumField = sumField
        self.meanField = meanField
        self.stdDevField = stdDevField

class BandAttPercentiles:
    """ This is passed to the populateRATWithPercentiles function """
    def __init__(self, percentile, fieldName):
        self.percentile = percentile
        self.fieldName = fieldName

class ShapeIndex:
    """ This is passed to the calcShapeIndices function """
    def __init__(self, colName, idx, colIdx=0):
        self.colName = colName
        self.colIdx = colIdx
        self.idx = idx

def exportCols2GDALImage(clumps, outimage, gdalformat, datatype, fields, ratband=1, tempDIR=None):
    """Exports columns of the raster attribute table as bands in a GDAL image. Utility function, exports each column individually then stacks them.

Where:

:param clumps: is a string containing the name of the input image file with RAT
:param outimage: is a string containing the name of the output gdal file
:param gdalformat: is a string containing the GDAL format for the output file - eg 'KEA'
:param datatype: is an int containing one of the values from rsgislib.TYPE_*
:param field: is a list of strings, providing the names of the column to be exported.
:param ratband: is an optional (default = 1) integer parameter specifying the image band to which the RAT is associated.

Example::

   clumps='./RATS/injune_p142_casi_sub_utm_clumps_elim_final_clumps_elim_final.kea'
   outimage='./TestOutputs/RasterGIS/injune_p142_casi_rgb_export.kea'
   gdalformat = 'KEA'
   datatype = rsgislib.TYPE_32FLOAT
   fields = ['RedAvg','GreenAvg','BlueAvg']
   rastergis.exportCols2GDALImage(clumps, outimage, gdalformat, datatype, fields)

"""
    import os
    import rsgislib
    from rsgislib import imageutils

    rsgisUtils = rsgislib.RSGISPyUtils()

    if tempDIR is None:
        tempDIR = os.path.split(outimage)[0]

    outExt=os.path.splitext(outimage)[-1]
    tempFileList = []

    # Export each field
    for field in fields:
        print('Exporting: ' + field)
        outTempFile = os.path.join(tempDIR, field + outExt)
        exportCol2GDALImage(clumps, outTempFile, gdalformat, datatype, field, ratband=1)
        tempFileList.append(outTempFile)

    # Stack Bands
    print('Stacking Bands')
    imageutils.stackImageBands(tempFileList, fields, outimage, None, 0, gdalformat, datatype)

    # Remove temp files
    print('Removing temp files')
    for tempFile in tempFileList:
        rsgisUtils.deleteFileWithBasename(tempFile)


def getRATLength(clumps_img, rat_band=1):
    """
A function which returns the length (i.e., number of rows) within the RAT.

:param clumps_img: path to the image file with the RAT
:param rat_band: the band within the image file for which the RAT is to read.

"""
    # Open input image file
    clumps_img_ds = gdal.Open(clumps_img, gdal.GA_ReadOnly)
    if clumps_img_ds is None:
        raise Exception("Could not open the inputted clumps image.")
    
    clumps_img_band = clumps_img_ds.GetRasterBand(rat_band)
    if clumps_img_band is None:
        raise Exception("Could not open the inputted clumps image band.")
    
    clumps_img_rat = clumps_img_band.GetDefaultRAT()
    if clumps_img_rat is None:
        raise Exception("Could not open the inputted clumps image band RAT.")
    
    nrows = clumps_img_rat.GetRowCount()
    
    clumps_img_ds = None
    return nrows


def getRATColumns(clumps_img, rat_band=1):
    """
A function which returns a list of column names within the RAT.

:param clumps_img: path to the image file with the RAT
:param rat_band: the band within the image file for which the RAT is to read.
:returns: list of column names.

"""
    # Open input image file
    clumps_img_ds = gdal.Open(clumps_img, gdal.GA_ReadOnly)
    if clumps_img_ds is None:
        raise Exception("Could not open the inputted clumps image.")
    
    clumps_img_band = clumps_img_ds.GetRasterBand(rat_band)
    if clumps_img_band is None:
        raise Exception("Could not open the inputted clumps image band.")
    
    clumps_img_rat = clumps_img_band.GetDefaultRAT()
    if clumps_img_rat is None:
        raise Exception("Could not open the inputted clumps image band RAT.")
        
    ncols = clumps_img_rat.GetColumnCount()
    col_names = []
    for col_idx in range(ncols):
        col_names.append(clumps_img_rat.GetNameOfCol(col_idx))
    
    clumps_img_ds = None
    return col_names


def getRATColumnsInfo(clumps_img, rat_band=1):
    """
A function which returns a dictionary of column names with type (GFT_Integer, GFT_Real, GFT_String)
and usage (e.g., GFU_Generic, GFU_PixelCount, GFU_Name, etc.) within the RAT.

:param clumps_img: path to the image file with the RAT
:param rat_band: the band within the image file for which the RAT is to read.
:returns: dict of column information.

"""
    # Open input image file
    clumps_img_ds = gdal.Open(clumps_img, gdal.GA_ReadOnly)
    if clumps_img_ds is None:
        raise Exception("Could not open the inputted clumps image.")

    clumps_img_band = clumps_img_ds.GetRasterBand(rat_band)
    if clumps_img_band is None:
        raise Exception("Could not open the inputted clumps image band.")

    clumps_img_rat = clumps_img_band.GetDefaultRAT()
    if clumps_img_rat is None:
        raise Exception("Could not open the inputted clumps image band RAT.")

    ncols = clumps_img_rat.GetColumnCount()
    col_info = dict()
    for col_idx in range(ncols):
        col_name = clumps_img_rat.GetNameOfCol(col_idx)
        col_type = clumps_img_rat.GetTypeOfCol(col_idx)
        col_usage = clumps_img_rat.GetUsageOfCol(col_idx)
        col_info[col_name] = dict()
        col_info[col_name]['type'] = col_type
        col_info[col_name]['usage'] = col_usage

    clumps_img_ds = None
    return col_info


def readRATNeighbours(clumps_img, start_row=None, end_row=None, rat_band=1):
    """
A function which returns a list of clumps neighbours from a KEA RAT. Note, the
neighbours are popualted using the function rsgislib.rastergis.findNeighbours. 
By default the whole datasets of neightbours is read to memory but the start_row 
and end_row variables can be used to read a subset of the RAT.

:param clumps_img: path to the image file with the RAT
:param start_row: the row within the RAT to start reading, if None will start at 0 (Default: None).
:param end_row: the row within the RAT to end reading, if None will end at n_rows within the RAT. (Default: None)
:param rat_band: the band within the image file for which the RAT is to read.
:returns: list of lists with neighbour indexes.

"""
    if not haveHDF5:
        raise Exception("Need the h5py library for this function")
       
    # Check that 'NumNeighbours' column exists
    rat_columns = getRATColumns(clumps_img, rat_band)
    if 'NumNeighbours' not in rat_columns:
        raise Exception("Clumps image RAT does not contain 'NumNeighbours' column - have you populated neightbours?")
    
    n_rows = getRATLength(clumps_img)
    
    if start_row is None:
        start_row = 0
    
    if end_row is None:
        end_row = n_rows
    
    clumps_h5_file = h5py.File(clumps_img)
    neighbours_path = 'BAND{}/ATT/NEIGHBOURS/NEIGHBOURS'.format(rat_band)
    neighbours = clumps_h5_file[neighbours_path]
    neighbours_data = neighbours[start_row:end_row]
    clumps_h5_file = None
    return neighbours_data






#! /usr/bin/env python

############################################################################
# Copyright (c) 2016 Dr. Peter Bunting, Aberystwyth University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# Purpose:  A series of classes to generate image plots from the ptxt text 
#           file format exported for plotting from the rsgis library. 
#
# Author: Pete Bunting
# Email: pfb@aber.ac.uk
# Date: 14/04/16
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
#############################################################################

import argparse
import csv
import numpy
import os

def readCSVColumns(csvFilePath, imgCol, grdCol, imgColVals, grdColVals):
    with open(csvFilePath) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            imgColVals.append(row[imgCol])
            grdColVals.append(row[grdCol])

def exportCSVFile(outFile, errMatrix, producerAcc, userAcc, producerAccCount, userAccCount, overallAcc, overallCorrCount, kappa, uniqueClasses):
    with open(outFile, 'wt') as csvfile:
        accWriter = csv.writer(csvfile, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
        
        accWriter.writerow(['Overall Accuracy (%)', round(overallAcc,2)])
        accWriter.writerow(['kappa', round(kappa,2)])
        
        accWriter.writerow([])
        accWriter.writerow(['Counts:'])
        
        numClasses = len(uniqueClasses)
        colNames = [' ']
        for i in range(numClasses):
            colNames.append(uniqueClasses[i])
        colNames.append('User')
        accWriter.writerow(colNames)
        
        for i in range(numClasses):
            row = []
            row.append(uniqueClasses[i])
            for j in range(numClasses):
                row.append(errMatrix[i,j])
            row.append(round(userAccCount[i],2))
            accWriter.writerow(row)
        
        prodRow = ['Producer']
        for i in range(numClasses):
            prodRow.append(round(producerAccCount[i],2))
        prodRow.append(overallCorrCount)
        accWriter.writerow(prodRow)
        
        accWriter.writerow([])
        accWriter.writerow(['Percentage:'])
        
        errMatrixPercent = (errMatrix / numpy.sum(errMatrix)) * 100
        
        colNames = [' ']
        for i in range(numClasses):
            colNames.append(uniqueClasses[i])
        colNames.append('User (%)')
        accWriter.writerow(colNames)
        
        for i in range(numClasses):
            row = []
            row.append(uniqueClasses[i])
            for j in range(numClasses):
                row.append(round(errMatrixPercent[i,j],2))
            row.append(round(userAcc[i],2))
            accWriter.writerow(row)
        
        prodRow = ['Producer (%)']
        for i in range(numClasses):
            prodRow.append(round(producerAcc[i],2))
        prodRow.append(round(overallAcc,2))
        accWriter.writerow(prodRow)


def exportToTex(outFile, errMatrix, producerAcc, userAcc, overallAcc, kappa, uniqueClasses):
    # Open File
    outputFile = open(outFile, 'w')
    
    for i in range(len(uniqueClasses)):
        uniqueClasses[i] = uniqueClasses[i].replace("_", "\_")
    
    # Write latex header to file
    outputFile.write('\\documentclass[12pt]{amsart}\n')
    outputFile.write('\\usepackage{geometry}\n')
    outputFile.write('\\geometry{a4paper}\n')
    outputFile.write('\\geometry{landscape}\n')
    outputFile.write('\\usepackage{graphicx}\n')
    outputFile.write('\\usepackage{amssymb}\n')
    outputFile.write('\\usepackage{epstopdf}\n')
    outputFile.write('\\usepackage{colortbl}\n')
    outputFile.write('\\usepackage[table]{xcolor}\n')
    outputFile.write('\\begin{document}\n')
    # Create the latex table
    outputFile.write('\\begin{table}[htbp]\n')
    caption = str("\\caption{Error Matrix. Overall Accuracy: ") + str(round(overallAcc,2)) + str("\% Kappa: ") + str(round((kappa),3)) +  str("}\n")
    outputFile.write(caption)
    outputFile.write('\\begin{center}\n')
    tabularLine = '\\begin{tabular}{|c|'
    for classVal in uniqueClasses:
        tabularLine =  tabularLine + 'c|'
    tabularLine =  tabularLine + 'c|}\n'
    outputFile.write(tabularLine)
    outputFile.write('\\hline\n')
    
    # Create table column titles
    titleLine = ''
    for classVal in uniqueClasses:
        titleLine =  titleLine + '&\\cellcolor{gray!50}\\textbf{' + str(classVal) + '}'
    titleLine =  titleLine + '&\\cellcolor{gray!50}\\textbf{User}\\\\\n'
    outputFile.write(titleLine)
    outputFile.write('\\hline\n')
    
    # Create main table body
    yIdx = 0
    for classValY in uniqueClasses:
        line = '\\cellcolor{gray!50}\\textbf{' + str(classValY) + '}'
        xIdx = 0
        for classValX in uniqueClasses:
            if classValY == classValX:
                line = line + '&\\cellcolor{red!50}' + str(errMatrix[yIdx,xIdx])
            else:
                line = line + '&' + str(errMatrix[yIdx,xIdx])
            xIdx = xIdx + 1    
        line = line + '&\\textbf{' + str(round(userAcc[yIdx],2)) + '\%}\\\\\n'
        yIdx = yIdx + 1
        outputFile.write(line)
        outputFile.write('\\hline\n')
    
    # Add bottom (producers) row
    line = '\\cellcolor{gray!50}\\textbf{Prod}'
    for val in producerAcc:
        line = line + '&\\textbf{' + str(round(val,2)) + '\%}'
    line = line + '&\\textbf{' + str(round(overallAcc,2)) + '\%}\\\\\n'
    outputFile.write(line)
    outputFile.write('\\hline\n')
    
    # End the table and document
    outputFile.write('\\end{tabular}\n')
    outputFile.write('\\end{center}\n')
    outputFile.write('\\label{tab:ErrMatrix}\n')
    outputFile.write('\\end{table}\n')
    outputFile.write('\\end{document}\n')
    
    # Close and flush output file
    outputFile.flush()
    outputFile.close()

def buildMatrix(imgColVals, grdColVals, uniqueElems):
    """
    build matrix
    """
    numClasses = len(uniqueElems)
    numFeats = len(imgColVals)
    errMatrix = numpy.zeros((numClasses,numClasses), dtype=numpy.float)
    
    for i in range(numFeats):
        imgClass = imgColVals[i]
        imgClassIdx = uniqueElems.index(imgClass)
        grdClass = grdColVals[i]
        grdClassIdx = uniqueElems.index(grdClass)
        #print("[" + imgClass + ", " + grdClass + "] = [ " + str(imgClassIdx) + ", " + str(grdClassIdx) + "]")
        errMatrix[imgClassIdx,grdClassIdx] = errMatrix[imgClassIdx,grdClassIdx] + 1
    
    producerAcc = numpy.zeros(numClasses, dtype=numpy.float)
    userAcc = numpy.zeros(numClasses, dtype=numpy.float)
    producerAccCount = numpy.zeros(numClasses, dtype=numpy.float)
    userAccCount = numpy.zeros(numClasses, dtype=numpy.float)
    overallCorrCount = 0.0
    
    for i in range(numClasses):
        corVal = float(errMatrix[i,i])
        sumRow = float(numpy.sum(errMatrix[i,]))
        sumCol = float(numpy.sum(errMatrix[...,i]))
        overallCorrCount = overallCorrCount + corVal

        if sumRow == 0:
            userAcc[i] = 0
            userAccCount[i] = 0
        else:
            userAcc[i] = corVal / sumRow
            userAccCount[i] = sumRow

        if sumCol == 0:
            producerAcc[i] = 0
            producerAccCount[i] = 0
        else:
            producerAcc[i] = corVal / sumCol
            producerAccCount[i] = sumCol
    
    overallAcc = (overallCorrCount / numpy.sum(errMatrix)) * 100
    producerAcc = producerAcc * 100
    userAcc = userAcc * 100
    
    kappaPartA = overallCorrCount * numpy.sum(producerAccCount)
    kappaPartB = numpy.sum(userAccCount * producerAccCount)
    kappaPartC = numpy.sum(errMatrix) * numpy.sum(errMatrix)

    kappa = float(kappaPartA - kappaPartB) / float(kappaPartC - kappaPartB)
    
    return errMatrix, producerAcc, userAcc, producerAccCount, userAccCount, overallAcc, overallCorrCount, kappa

def produceErrorMatrix(inputCSVs, outputCSVFile, outputTEXFile, outTex, outPDF, imgCol, grdCol):
    """
    Read data, calc and output.
    """
    imgColVals = []
    grdColVals = []
    
    for csvFile in inputCSVs:
        readCSVColumns(csvFile, imgCol, grdCol, imgColVals, grdColVals)
        
    #print(imgColVals)
    #print(grdColVals)
    
    allVals = imgColVals + grdColVals
    
    uniqueClasses = list(set(allVals))
    print(uniqueClasses)
    
    errMatrix, producerAcc, userAcc, producerAccCount, userAccCount, overallAcc, overallCorrCount, kappa = buildMatrix(imgColVals, grdColVals, uniqueClasses)
        
    exportCSVFile(outputCSVFile, errMatrix, producerAcc, userAcc, producerAccCount, userAccCount, overallAcc, overallCorrCount, kappa, uniqueClasses)
    
    if outTex:
        exportToTex(outputTEXFile, errMatrix, producerAcc, userAcc, overallAcc, kappa, uniqueClasses)
        if outPDF:
            cmd = "pdflatex " + outputTEXFile
            os.system(cmd)

if __name__ == '__main__':
    """
    The command line user interface
    """
    parser = argparse.ArgumentParser(prog='rsgiscalcaccmatrix.py', description='''Calculate Error Matrix''')

    parser.add_argument("-c", "--csv", type=str, required=True, help='''Output CSV File.''')
    parser.add_argument("-t", "--tex", type=str, help='''Output TEX File.''')    
    parser.add_argument("-p", "--pdf", action='store_true', default=False, help='''Output PDF File.''')
    parser.add_argument("-i", "--inputs", type=str, required=True, nargs='*', help='''Input CSV files''')    
    parser.add_argument("-r", "--imgcol", type=str, required=True, help='''Image Column Name''')
    parser.add_argument("-g", "--grdcol", type=str, required=True, help='''Ground Truth Column Name ''')
    
    # Call the parser to parse the arguments.
    args = parser.parse_args()
    
    outTex = True
    if args.tex == None:
        outTex = False        
    
    produceErrorMatrix(args.inputs, args.csv, args.tex, outTex, args.pdf, args.imgcol, args.grdcol)
    
    



#!/usr/bin/env python

#############################################
# rsgislibcalcindices.py
#
#  Copyright 2015 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose:  Calculate useful indices quickly and easily.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 20/02/2015
# Version: 1.0
# 
#############################################
import sys, argparse

try:
    from rsgislib import imagecalc
except ImportError:
    print("ERROR: Couldn't import RSGISLib python modules")
    sys.exit()

if __name__ == '__main__':    
    parser = argparse.ArgumentParser(prog='rsgislibcalcindices.py')

    parser.add_argument("-i", "--input", type=str, help='''Specify the input image file.''')
    parser.add_argument("-o", "--output", type=str,help='''Specify the output image file.''')
    parser.add_argument("-c", "--calc", type=str, choices=['ndvi', 'wbi'], help='''Specify the indice to calculate (nvdi, wbi)''')
    parser.add_argument("-s", "--stats", action='store_true', default=False, help='''Specifies that the image statistics and pyramids should be build for all output images.''')
    parser.add_argument("-b", "--bands", type=int, nargs='+', help='''Specify the image bands for the index (NDVI: Red, NIR. WBI: Blue, NIR)''')
    parser.add_argument("-f", "--format", type=str, default='KEA', help='''Specify the output image file format (Default: KEA).''')
    # Call the parser to parse the arguments.
    args = parser.parse_args()
        

    if args.calc == "ndvi":
        imagecalc.calcNDVI(args.input, args.bands[0], args.bands[1], args.output, args.stats, args.format)
    elif args.calc == "wbi":
        imagecalc.calcWBI(args.input, args.bands[0], args.bands[1], args.output, args.stats, args.format)
    else:
        print("Don't recongise indice specified.")
        
        
        
        
        
